package main

// this solution uses unbuffered channels

import (
	"fmt"
	"math/rand"
	"sync"
	"time"
)

// Philosopher represents a philosopher with a name and left/right forks.
type Philosopher struct {
	name       string
	leftFork   chan struct{}
	rightFork  chan struct{}
}

// dine handles the eating and thinking routine for each philosopher.
func (p *Philosopher) dine(wg *sync.WaitGroup) {
	defer wg.Done()
	for i := 0; i < 7; i++ { // each philosopher will eat/think 7 times
		p.think()
		p.eat()
	}
	fmt.Printf("%s is done being.\n", p.name)
}

// think simulates thinking time.
func (p *Philosopher) think() {
	fmt.Printf("%s is thinking...\n", p.name)
	time.Sleep(time.Duration(rand.Intn(8000)) * time.Millisecond)
}

// eat simulates eating by picking up and releasing forks.
func (p *Philosopher) eat() {
	// Pick up forks (unbuffered channels require immediate receiving).
	fmt.Printf("%s is hungry and tries to pick up forks.\n", p.name)
	<-p.leftFork
	<-p.rightFork
	fmt.Printf("%s picked up forks and is eating.\n", p.name)
	time.Sleep(time.Duration(rand.Intn(8000)) * time.Millisecond)

	// Release forks after eating.
	p.leftFork <- struct{}{}
	p.rightFork <- struct{}{}
	fmt.Printf("%s finished eating and put down forks.\n", p.name)
}

func main() {
	rand.Seed(time.Now().UnixNano())
	numPhilosophers := 5
	var wg sync.WaitGroup
	wg.Add(numPhilosophers)

	// Initialize forks (one fork per philosopher) as unbuffered channels.
	forks := make([]chan struct{}, numPhilosophers)
	for i := 0; i < numPhilosophers; i++ {
		forks[i] = make(chan struct{}) // Unbuffered channel
		go func(fork chan struct{}) {
			fork <- struct{}{} // Initialize each fork as available
		}(forks[i])
	}

	// Create philosophers and assign them forks.
	philosophers := make([]*Philosopher, numPhilosophers)
	for i := 0; i < numPhilosophers; i++ {
		leftFork := forks[i]
		rightFork := forks[(i+1)%numPhilosophers]
		philosophers[i] = &Philosopher{
			name:      fmt.Sprintf("Philosopher %d", i+1),
			leftFork:  leftFork,
			rightFork: rightFork,
		}
		go philosophers[i].dine(&wg)
	}

	// make main wait for all philosophers to finish dining.
	wg.Wait()
	fmt.Println("All philosophers are done dining.")
}
